/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package fr.loof.fonzie.dialects;

import java.lang.reflect.Field;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.persistence.Id;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.ReflectionUtils;

public abstract class JPADialectSupport
    implements JPADialect
{
    protected static Logger logger = LoggerFactory.getLogger( JPADialectSupport.class );

    protected Object readIdentifier( Object entity )
    {
        Field f = getIdField( entity );
        logger.trace( "read Id field {} from entity {}", f, entity );
        return ReflectionUtils.getField( f, entity );
    }

    @SuppressWarnings( "unchecked" )
    private static final ConcurrentMap<Class, Field> ID_FIELDS = new ConcurrentHashMap<Class, Field>();

    /**
     * Detects the @Id attribute of an entity and return the {@link Field} object that can be used to access it.
     * <p>
     * Internally uses a ConcurrentMap for better performances.
     */
    public static Field getIdField( Object entity )
    {
        Class<?> targetClass = entity.getClass();
        Field f = ID_FIELDS.get( targetClass );
        if ( f == null )
        {
            f = findIdField( targetClass );
            logger.debug( "{} ID field is {}", targetClass, f );
            ID_FIELDS.putIfAbsent( targetClass, f );
        }
        return f;
    }

    /**
     * Fiel @Id {@link Field} in the targetClass Hierarchy
     */
    private static Field findIdField( Class<?> targetClass )
    {
        do
        {
            for ( Field field : targetClass.getDeclaredFields() )
            {
                if ( field.getAnnotation( Id.class ) != null )
                {
                    ReflectionUtils.makeAccessible( field );
                    return field;
                }
            }
            targetClass = targetClass.getSuperclass();
        }
        while ( targetClass != null && targetClass != Object.class );
        throw new IllegalArgumentException( "Entity " + targetClass + " has no @Id field" );
    }

}
